import tqdm
import tqdm.auto
import random
import numpy as np
from joblib import Parallel, delayed
from multiprocessing import cpu_count
from fastdtw import fastdtw
from collections import Counter
from scipy.spatial.distance import euclidean
from sklearn.metrics import confusion_matrix
from ...internals import _Validator

class DTWKNN:
    """A k-Nearest Neighbor classifier that compares differing length observation sequences using the efficient FastDTW dynamic time warping algorithm.

    Parameters
    ----------
    k: int
        Number of neighbors.

    radius: int
        Radius parameter for FastDTW.

        **See**: `Stan Salvador, and Philip Chan. "FastDTW: Toward accurate dynamic time warping in linear time and space." Intelligent Data Analysis 11.5 (2007), 561-580. <https://pdfs.semanticscholar.org/05a2/0cde15e172fc82f32774dd0cf4fe5827cad2.pdf>`_

    metric: callable
        Distance metric for FastDTW.
    """

    def __init__(self, k, radius, metric=euclidean):
        self._val = _Validator()
        self._k = self._val.restricted_integer(
            k, lambda x: x > 0, desc='number of neighbors', expected='greater than zero')
        self._radius = self._val.restricted_integer(
            radius, lambda x: x > 0, desc='radius parameter', expected='greater than zero')
        self._metric = metric

    def fit(self, X, y):
        """Fits the classifier by adding labeled training observation sequences.

        Parameters
        ----------
        X: List[numpy.ndarray]
            A list of multiple observation sequences.

        y: List[str]
            A list of labels for the observation sequences.
        """
        self._X, self._y = self._val.observation_sequences_and_labels(X, y)

    def predict(self, X, verbose=True, n_jobs=1):
        """Predicts the label for an observation sequence (or multiple sequences).

        Parameters
        ----------
        X: numpy.ndarray or List[numpy.ndarray]
            An individual observation sequence or a list of multiple observation sequences.

        verbose: bool
            Whether to display a progress bar or not.

        n_jobs: int
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        prediction(s): str or List[str]
            The predicted label(s) for the observation sequence(s).
        """
        try:
            (self._X, self._y)
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted before predictions are made')

        self._val.observation_sequences(X, allow_single=True)
        self._val.boolean(verbose, desc='verbose')
        self._val.restricted_integer(n_jobs, lambda x: x == -1 or x > 0, 'number of jobs', '-1 or greater than zero')

        # FastDTW distance measure
        distance = lambda x1, x2: fastdtw(x1, x2, radius=self._radius, dist=self._metric)[0]

        def find_modes(distances):
            idx = np.argpartition(distances, self._k)[:self._k]
            neighbor_labels = [self._y[i] for i in idx]
            # Find the modal labels
            counter = Counter(neighbor_labels)
            max_count = max(counter.values())
            return [k for k, v in counter.items() if v == max_count]

        if isinstance(X, np.ndarray):
            distances = [distance(X, x) for x in tqdm.auto.tqdm(self._X, desc='Calculating distances', disable=not(verbose))]
            modes = find_modes(distances)
            # Randomly select one of the modal labels
            return random.choice(modes)
        else:
            if n_jobs == 1:
                labels = []
                for O in tqdm.auto.tqdm(X, desc='Classifying examples', disable=not(verbose)):
                    distances = [distance(O, x) for x in self._X]
                    modes = find_modes(distances)
                    # Randomly select one of the modal labels
                    labels.append(random.choice(modes))
                return labels
            else:
                def parallel_predict(process, X_chunk):
                    labels = []
                    for O in tqdm.tqdm(X_chunk, desc='Classifying examples (process {})'.format(process), disable=not(verbose), position=process-1):
                        distances = [distance(O, x) for x in self._X]
                        modes = find_modes(distances)
                        labels.append(random.choice(modes))
                    return labels

                n_jobs = cpu_count() if n_jobs == -1 else n_jobs
                X_chunks = [list(chunk) for chunk in np.array_split(X, n_jobs)]
                labels = Parallel(n_jobs=n_jobs)(delayed(parallel_predict)(i+1, chunk) for i, chunk in enumerate(X_chunks))
                return [label for sublist in labels for label in sublist] # Flatten the resulting array

    def evaluate(self, X, y, labels=None, verbose=True, n_jobs=1):
        """Evaluates the performance of the classifier on a batch of observation sequences and their labels.

        Parameters
        ----------
        X: List[numpy.ndarray]
            A list of multiple observation sequences.

        y: List[str]
            A list of labels for the observation sequences.

        labels: List[str]
            A list of labels for ordering the axes of the confusion matrix.

        verbose: bool
            Whether to display a progress bar for predictions or not.

        n_jobs: int
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        accuracy: float
            The categorical accuracy of the classifier on the observation sequences.

        confusion: numpy.ndarray
            The confusion matrix representing the discrepancy between predicted and actual labels.
        """
        self._val.observation_sequences_and_labels(X, y)
        self._val.boolean(verbose, desc='verbose')

        if labels is not None:
            self._val.list_of_strings(labels, desc='confusion matrix labels')

        # Classify each observation sequence and calculate confusion matrix
        predictions = self.predict(X, verbose=verbose, n_jobs=n_jobs)
        cm = confusion_matrix(y, predictions, labels=labels)

        return np.sum(np.diag(cm)) / np.sum(cm), cm


import scipy.fftpack
import numpy as np
from ..internals import _Validator

def trim_zeros(X):
    """Trim zero-observations from the input observation sequence(s).

    Parameters
    ----------
    X: numpy.ndarray or List[numpy.ndarray]
        An individual observation sequence or a list of multiple observation sequences.

    Returns
    -------
    trimmed: numpy.ndarray or List[numpy.ndarray]
        The zero-trimmed input observation sequence(s).
    """
    val = _Validator()
    val.observation_sequences(X, allow_single=True)
    return _trim_zeros(X)

def _trim_zeros(X):
    def transform(x):
        return x[~np.all(x == 0, axis=1)]

    if isinstance(X, list):
        return [transform(x) for x in X]
    elif isinstance(X, np.ndarray):
        return transform(X)

def center(X):
    """Centers an observation sequence (or multiple sequences) by centering observations around the mean.

    Parameters
    ----------
    X: numpy.ndarray or List[numpy.ndarray]
        An individual observation sequence or a list of multiple observation sequences.

    Returns
    -------
    centered: numpy.ndarray or List[numpy.ndarray]
        The centered input observation sequence(s).
    """
    val = _Validator()
    val.observation_sequences(X, allow_single=True)
    return _center(X)

def _center(X):
    def transform(x):
        return x - x.mean(axis=0)

    if isinstance(X, list):
        return [transform(x) for x in X]
    elif isinstance(X, np.ndarray):
        return transform(X)

def standardize(X):
    """Standardizes an observation sequence (or multiple sequences) by transforming observations
    so that they have zero mean and unit variance.

    Parameters
    ----------
    X: numpy.ndarray or List[numpy.ndarray]
        An individual observation sequence or a list of multiple observation sequences.

    Returns
    -------
    standardized: numpy.ndarray or List[numpy.ndarray]
        The standardized input observation sequence(s).
    """
    val = _Validator()
    val.observation_sequences(X, allow_single=True)
    return _standardize(X)

def _standardize(X):
    def transform(x):
        return (x - x.mean(axis=0)) / x.std(axis=0)

    if isinstance(X, list):
        return [transform(x) for x in X]
    elif isinstance(X, np.ndarray):
        return transform(X)

def downsample(X, n, method='decimate'):
    """Downsamples an observation sequence (or multiple sequences) by either:

    - Decimating the next :math:`n-1` observations
    - Averaging the current observation with the next :math:`n-1` observations

    Parameters
    ----------
    X: numpy.ndarray or List[numpy.ndarray]
        An individual observation sequence or a list of multiple observation sequences.

    n: int
        Downsample factor.

    method: {'decimate', 'average'}
        The downsampling method.

    Returns
    -------
    downsampled: numpy.ndarray or List[numpy.ndarray]
        The downsampled input observation sequence(s).
    """
    val = _Validator()
    val.observation_sequences(X, allow_single=True)
    val.restricted_integer(n, lambda x: x > 1, desc='downsample factor', expected='greater than one')
    val.one_of(method, ['decimate', 'average'], desc='downsampling method')

    if isinstance(X, np.ndarray):
        val.restricted_integer(n, lambda x: x <= len(X),
            desc='downsample factor', expected='no greater than the number of frames')
    else:
        val.restricted_integer(n, lambda x: x <= min(len(x) for x in X),
            desc='downsample factor', expected='no greater than the number of frames in the shortest sequence')

    return _downsample(X, n, method)

def _downsample(X, n, method):
    def transform(x):
        N, D = x.shape
        if method == 'decimate':
            return np.delete(x, [i for i in range(N) if i % n != 0], 0)
        elif method == 'average':
            r = len(x) % n
            xn, xr = (x, None) if r == 0 else (x[:-r], x[-r:])
            dxn = xn.T.reshape(-1, n).mean(axis=1).reshape(D, -1).T
            return dxn if xr is None else np.vstack((dxn, xr.mean(axis=0)))

    if isinstance(X, list):
        return [transform(x) for x in X]
    elif isinstance(X, np.ndarray):
        return transform(X)

def fft(X):
    """Applies a Discrete Fourier Transform to the input observation sequence(s).

    Parameters
    ----------
    X: numpy.ndarray or List[numpy.ndarray]
        An individual observation sequence or a list of multiple observation sequences.

    Returns
    -------
    transformed: numpy.ndarray or List[numpy.ndarray]
        The transformed input observation sequence(s).
    """
    val = _Validator()
    val.observation_sequences(X, allow_single=True)
    return _fft(X)

def _fft(X):
    def transform(x):
        return scipy.fftpack.rfft(x, axis=0)

    if isinstance(X, list):
        return [transform(x) for x in X]
    elif isinstance(X, np.ndarray):
        return transform(X)

def filtrate(X, n, method='median'):
    """Applies a median or mean filter to the input observation sequence(s).

    Parameters
    ----------
    X: numpy.ndarray or List[numpy.ndarray]
        An individual observation sequence or a list of multiple observation sequences.

    n: int
        Window size.

    method: {'median', 'mean'}
        The filtering method.

    Returns
    -------
    filtered: numpy.ndarray or List[numpy.ndarray]
        The filtered input observation sequence(s).
    """
    val = _Validator()
    val.observation_sequences(X, allow_single=True)
    val.restricted_integer(n, lambda x: x > 1, desc='window size', expected='greater than one')
    val.one_of(method, ['median', 'mean'], desc='filtering method')

    if isinstance(X, np.ndarray):
        val.restricted_integer(n, lambda x: x <= len(X),
            desc='window size', expected='no greater than the number of frames')
    else:
        val.restricted_integer(n, lambda x: x <= min(len(x) for x in X),
            desc='window size', expected='no greater than the number of frames in the shortest sequence')

    return _filtrate(X, n, method)

def _filtrate(X, n, method):
    def transform(x):
        measure = np.median if method == 'median' else np.mean
        filtered = []
        right = n // 2
        left = (n - 1) - right
        for i in range(len(x)):
            l, m, r = x[((i - left) * (left < i)):i], x[i], x[(i + 1):(i + 1 + right)]
            filtered.append(measure(np.vstack((l, m, r)), axis=0))
        return np.array(filtered)

    if isinstance(X, list):
        return [transform(x) for x in X]
    elif isinstance(X, np.ndarray):
        return transform(X)


import numpy as np
from .topology import _Topology

class _LeftRightTopology(_Topology):
    """Represents the topology for a left-right HMM, imposing an upper-triangular transition matrix.

    Parameters
    ----------
    n_states: int
        Number of states in the HMM.

    random_state: numpy.random.RandomState
        A random state object for reproducible randomness.
    """

    def __init__(self, n_states: int, random_state: np.random.RandomState):
        super().__init__(n_states, random_state)

    def uniform_transitions(self) -> np.ndarray:
        """Sets the transition matrix as uniform (equal probability of transitioning
            to all other possible states from each state) corresponding to the topology.

        Returns
        -------
        transitions: numpy.ndarray
            The uniform transition matrix of shape `(n_states, n_states)`.
        """
        upper_ones = np.triu(np.ones((self._n_states, self._n_states)))
        upper_divisors = np.triu(np.tile(np.arange(self._n_states, 0, -1), (self._n_states, 1)).T)
        lower_ones = np.tril(np.ones(self._n_states), k=-1) # One-pad lower triangle to prevent zero division
        return upper_ones / (upper_divisors + lower_ones)

    def random_transitions(self) -> np.ndarray:
        """Sets the transition matrix as random (random probability of transitioning
        to all other possible states from each state) by sampling probabilities
        from a Dirichlet distribution - according to the topology.

        Parameters
        ----------
        transitions: numpy.ndarray
            The random transition matrix of shape `(n_states, n_states)`.
        """
        transitions = self._random_state.dirichlet(np.ones(self._n_states), size=self._n_states)
        lower_sums = np.sum(np.tril(transitions, k=-1), axis=1) # Amount to be redistributed per row
        quantities = np.arange(self._n_states, 0, -1) # Number of elements per row to redistribute evenly to
        upper_ones = np.triu(np.ones((self._n_states, self._n_states)))
        redist = (lower_sums / quantities).reshape(-1, 1) * upper_ones
        return np.triu(transitions) + redist

    def validate_transitions(self, transitions: np.ndarray) -> None:
        """Validates a transition matrix according to the topology's restrictions.

        Parameters
        ----------
        transitions: numpy.ndarray
            The transition matrix to validate.
        """
        super().validate_transitions(transitions)
        if not np.allclose(transitions, np.triu(transitions)):
            raise ValueError('Left-right transition matrix must be upper-triangular')


import pytest
import warnings
import numpy as np
from copy import deepcopy
with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=DeprecationWarning)
    from sequentia.classifiers import DTWKNN
from ....support import assert_equal, assert_not_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

labels = ['c0', 'c1', 'c2', 'c3', 'c4']

# Create some sample data
X = [rng.random((10 * i, 3)) for i in range(1, 7)]
y = ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']
x = X[0]

clfs = [
    DTWKNN(k=1, radius=1),
    DTWKNN(k=2, radius=5),
    DTWKNN(k=3, radius=10)
]

for clf in clfs:
    clf.fit(X, y)

# ============ #
# DTWKNN.fit() #
# ============ #

def test_fit_sets_attributes():
    """Check that fitting sets the hidden attributes"""
    clf = clfs[0]
    clf.fit(X, y)
    assert clf._X == X
    assert clf._y == y

# ================ #
# DTWKNN.predict() #
# ================ #

def test_predict_without_fit():
    """Predict without fitting the model"""
    with pytest.raises(RuntimeError) as e:
        DTWKNN(k=1, radius=1).predict(x, verbose=False)
    assert str(e.value) == 'The classifier needs to be fitted before predictions are made'

def test_predict_single_k1_r1_verbose(capsys):
    """Verbosely predict a single observation sequence (k=1, r=1)"""
    prediction = clfs[0].predict(x, verbose=True)
    assert 'Calculating distances' in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k1_r1_no_verbose(capsys):
    """Silently predict a single observation sequence (k=1, r=1)"""
    prediction = clfs[0].predict(x, verbose=False)
    assert 'Calculating distances' not in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k2_r5_verbose(capsys):
    """Verbosely predict a single observation sequence (k=2, r=5)"""
    prediction = clfs[1].predict(x, verbose=True)
    assert 'Calculating distances' in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k2_r5_no_verbose(capsys):
    """Silently predict a single observation sequence (k=2, r=5)"""
    prediction = clfs[1].predict(x, verbose=False)
    assert 'Calculating distances' not in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k3_r10_verbose(capsys):
    """Verbosely predict a single observation sequence (k=3, r=10)"""
    prediction = clfs[2].predict(x, verbose=True)
    assert 'Calculating distances' in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k3_r10_no_verbose(capsys):
    """Silently predict a single observation sequence (k=3, r=10)"""
    prediction = clfs[2].predict(x, verbose=False)
    assert 'Calculating distances' not in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_multiple_k1_r1_verbose(capsys):
    """Verbosely predict multiple observation sequences (k=1, r=1)"""
    predictions = clfs[0].predict(X, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert predictions == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']

def test_predict_multiple_k1_r1_no_verbose(capsys):
    """Silently predict multiple observation sequences (k=1, r=1)"""
    predictions = clfs[0].predict(X, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert predictions == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']

def test_predict_multiple_k2_r5_verbose(capsys):
    """Verbosely predict multiple observation sequences (k=2, r=5)"""
    predictions = clfs[1].predict(X, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert len(predictions) == 6

def test_predict_multiple_k2_r5_no_verbose(capsys):
    """Silently predict multiple observation sequences (k=2, r=5)"""
    predictions = clfs[1].predict(X, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert len(predictions) == 6

def test_predict_multiple_k3_r10_verbose(capsys):
    """Verbosely predict multiple observation sequences (k=3, r=10)"""
    predictions = clfs[2].predict(X, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert len(predictions) == 6

def test_predict_multiple_k3_r10_no_verbose(capsys):
    """Silently predict multiple observation sequences (k=3, r=10)"""
    predictions = clfs[2].predict(X, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert len(predictions) == 6

# ================= #
# DTWKNN.evaluate() #
# ================= #

def test_evaluate_with_labels_k1_r1_verbose(capsys):
    """Verbosely evaluate observation sequences with labels (k=1, r=1)"""
    acc, cm = clfs[0].evaluate(X, y, labels=labels, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)
    assert cm.shape == (5, 5)

def test_evaluate_with_labels_k1_r1_no_verbose(capsys):
    """Silently evaluate observation sequences with labels (k=1, r=1)"""
    acc, cm = clfs[0].evaluate(X, y, labels=labels, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)
    assert cm.shape == (5, 5)

def test_evaluate_with_no_labels_k1_r1_verbose(capsys):
    """Verbosely evaluate observation sequences without labels (k=1, r=1)"""
    acc, cm = clfs[0].evaluate(X, y, labels=None, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)

def test_evaluate_with_no_labels_k1_r1_no_verbose(capsys):
    """Silently evaluate observation sequences without labels (k=1, r=1)"""
    acc, cm = clfs[0].evaluate(X, y, labels=None, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)

def test_evaluate_with_labels_k2_r5_verbose(capsys):
    """Verbosely evaluate observation sequences with labels (k=2, r=5)"""
    acc, cm = clfs[1].evaluate(X, y, labels=labels, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)
    assert cm.shape == (5, 5)

def test_evaluate_with_labels_k2_r5_no_verbose(capsys):
    """Silently evaluate observation sequences with labels (k=2, r=5)"""
    acc, cm = clfs[1].evaluate(X, y, labels=labels, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)
    assert cm.shape == (5, 5)

def test_evaluate_with_no_labels_k2_r5_verbose(capsys):
    """Verbosely evaluate observation sequences without labels (k=2, r=5)"""
    acc, cm = clfs[1].evaluate(X, y, labels=None, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)

def test_evaluate_with_no_labels_k2_r5_no_verbose(capsys):
    """Silently evaluate observation sequences without labels (k=2, r=5)"""
    acc, cm = clfs[1].evaluate(X, y, labels=None, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)

def test_evaluate_with_labels_k3_r10_verbose(capsys):
    """Verbosely evaluate observation sequences with labels (k=3, r=10)"""
    acc, cm = clfs[2].evaluate(X, y, labels=labels, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)
    assert cm.shape == (5, 5)

def test_evaluate_with_labels_k3_r10_no_verbose(capsys):
    """Silently evaluate observation sequences with labels (k=3, r=10)"""
    acc, cm = clfs[2].evaluate(X, y, labels=labels, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)
    assert cm.shape == (5, 5)

def test_evaluate_with_no_labels_k3_r10_verbose(capsys):
    """Verbosely evaluate observation sequences without labels (k=3, r=10)"""
    acc, cm = clfs[2].evaluate(X, y, labels=None, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)

def test_evaluate_with_no_labels_k3_r10_no_verbose(capsys):
    """Silently evaluate observation sequences without labels (k=3, r=10)"""
    acc, cm = clfs[2].evaluate(X, y, labels=None, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)


import numpy as np
import pomegranate as pg
from .topologies.ergodic import _ErgodicTopology
from .topologies.left_right import _LeftRightTopology
from ...internals import _Validator

class HMM:
    """A hidden Markov model representing an isolated temporal sequence class.

    Parameters
    ----------
    label: str
        A label for the model, corresponding to the class being represented.

    n_states: int
        The number of states for the model.

    topology: {'ergodic', 'left-right'}
        The topology for the model.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.

    Attributes
    ----------
    label: str
        The label for the model.

    n_states: int
        The number of states for the model.

    n_seqs: int
        The number of observation sequences use to train the model.

    initial: numpy.ndarray
        The initial state distribution of the model.

    transitions: numpy.ndarray
        The transition matrix of the model.
    """

    def __init__(self, label, n_states, topology='left-right', random_state=None):
        self._val = _Validator()
        self._label = self._val.string(label, 'model label')
        self._n_states = self._val.restricted_integer(
            n_states, lambda x: x > 0, desc='number of states', expected='greater than zero')
        self._val.one_of(topology, ['ergodic', 'left-right'], desc='topology')
        self._random_state = self._val.random_state(random_state)

        if topology == 'ergodic':
            self._topology = _ErgodicTopology(self._n_states, self._random_state)
        elif topology == 'left-right':
            self._topology = _LeftRightTopology(self._n_states, self._random_state)

    def set_uniform_initial(self):
        """Sets a uniform initial state distribution."""
        self._initial = self._topology.uniform_initial()

    def set_random_initial(self):
        """Sets a random initial state distribution."""
        self._initial = self._topology.random_initial()

    def set_uniform_transitions(self):
        """Sets a uniform transition matrix according to the topology."""
        self._transitions = self._topology.uniform_transitions()

    def set_random_transitions(self):
        """Sets a random transition matrix according to the topology."""
        self._transitions = self._topology.random_transitions()

    def fit(self, X, n_jobs=1):
        """Fits the HMM to observation sequences assumed to be labeled as the class that the model represents.

        Parameters
        ----------
        X: List[numpy.ndarray]
            Collection of multivariate observation sequences, each of shape :math:`(T \\times D)` where
            :math:`T` may vary per observation sequence.

        n_jobs: int
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.
        """
        self._val.observation_sequences(X)
        self._val.restricted_integer(n_jobs, lambda x: x == -1 or x > 0, 'number of jobs', '-1 or greater than zero')

        try:
            (self._initial, self._transitions)
        except AttributeError as e:
            raise AttributeError('Must specify initial state distribution and transitions before the HMM can be fitted') from e

        self._n_seqs = len(X)
        self._n_features = X[0].shape[1]

        # Create a multivariate Gaussian emission distribution using combined samples for initial parameter esimation
        dist = pg.MultivariateGaussianDistribution.from_samples(np.concatenate(X))

        # Create the HMM object
        self._model = pg.HiddenMarkovModel.from_matrix(
            name=self._label,
            transition_probabilities=self._transitions,
            distributions=[dist.copy() for _ in range(self._n_states)],
            starts=self._initial
        )

        # Perform the Baum-Welch algorithm to fit the model to the observations
        self._model.fit(X, n_jobs=n_jobs)

        # Update the initial state distribution and transitions to reflect the updated parameters
        inner_tx = self._model.dense_transition_matrix()[:, :self._n_states]
        self._initial = inner_tx[self._n_states]
        self._transitions = inner_tx[:self._n_states]

    def forward(self, sequence):
        """Runs the forward algorithm to calculate the (negative log) likelihood of the model generating an observation sequence.

        Parameters
        ----------
        sequence: numpy.ndarray
            An individual sequence of observations of size :math:`(T \\times D)` where
            :math:`T` is the number of time frames (or observations) and
            :math:`D` is the number of features.

        Returns
        -------
        negative log-likelihood: float
            The negative log-likelihood of the model generating the observation sequence.
        """
        try:
            self._model
        except AttributeError as e:
            raise AttributeError('The model must be fitted before running the forward algorithm') from e

        if not isinstance(sequence, np.ndarray):
            raise TypeError('Sequence of observations must be a numpy.ndarray')
        if not sequence.ndim == 2:
            raise ValueError('Sequence of observations must be two-dimensional')
        if not sequence.shape[1] == self._n_features:
            raise ValueError('Number of observation features must match the dimensionality of the original data used to fit the model')

        return -self._model.log_probability(sequence)

    @property
    def label(self):
        return self._label

    @property
    def n_states(self):
        return self._n_states

    @property
    def n_seqs(self):
        try:
            return self._n_seqs
        except AttributeError as e:
            raise AttributeError('The model has not been fitted and has not seen any observation sequences') from e

    @property
    def initial(self):
        try:
            return self._initial
        except AttributeError as e:
            raise AttributeError('No initial state distribution has been defined') from e

    @initial.setter
    def initial(self, probabilities):
        self._topology.validate_initial(probabilities)
        self._initial = probabilities

    @property
    def transitions(self):
        try:
            return self._transitions
        except AttributeError as e:
            raise AttributeError('No transition matrix has been defined') from e

    @transitions.setter
    def transitions(self, probabilities):
        self._topology.validate_transitions(probabilities)
        self._transitions = probabilities


import pytest
import warnings
import numpy as np
from copy import deepcopy
from sequentia.classifiers import HMM, HMMClassifier
from ....support import assert_equal, assert_not_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

# Set of possible labels
labels = ['c{}'.format(i) for i in range(5)]

# Create and fit some sample HMMs
hmm_list, hmm_dict = [], {}
for i, label in enumerate(labels):
    hmm = HMM(label=label, n_states=(i + 3), random_state=rng)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit([np.arange((i + j * 20) * 30).reshape(-1, 3) for j in range(2, 5)])
    hmm_list.append(hmm)
    hmm_dict[label] = hmm

# Create some sample test data and labels
X = [np.arange((i + 2 * 20) * 30).reshape(-1, 3) for i in range(2, 5)]
Y = ['c0', 'c1', 'c1']
x, y = X[0], 'c1'

# Fit a classifier
hmm_clf = HMMClassifier()
hmm_clf.fit(hmm_list)

# =================== #
# HMMClassifier.fit() #
# =================== #

def test_fit_list():
    """Fit the classifier using a list of HMMs"""
    clf = HMMClassifier()
    clf.fit(hmm_list)
    assert clf._models == hmm_list

def test_fit_list_empty():
    """Fit the classifier using an empty list"""
    clf = HMMClassifier()
    with pytest.raises(RuntimeError) as e:
        clf.fit([])
    assert str(e.value) == 'Must fit the classifier with at least one HMM'

def test_fit_list_invalid():
    """Fit the classifier using a list of invalid types"""
    clf = HMMClassifier()
    with pytest.raises(TypeError) as e:
        clf.fit(['a', 'b'])
    assert str(e.value) == 'Expected all models to be HMM objects'

def test_fit_dict():
    """Fit the classifier using a dict of HMMs"""
    clf = HMMClassifier()
    clf.fit(hmm_dict)
    assert clf._models == list(hmm_dict.values())

def test_fit_dict_empty():
    """Fit the classifier using an empty dict"""
    clf = HMMClassifier()
    with pytest.raises(RuntimeError) as e:
        clf.fit({})
    assert str(e.value) == 'Must fit the classifier with at least one HMM'

def test_fit_dict_invalid():
    """Fit the classifier using a dict of invalid types"""
    clf = HMMClassifier()
    with pytest.raises(TypeError) as e:
        clf.fit({'a': 0, 'b': 0})
    assert str(e.value) == 'Expected all models to be HMM objects'

def test_fit_invalid():
    """Fit the classifier with an invalid"""
    clf = HMMClassifier()
    with pytest.raises(TypeError) as e:
        clf.fit(tuple(hmm_list))
    assert str(e.value) == 'Expected models to be a list or dict of HMM objects'

# ======================= #
# HMMClassifier.predict() #
# ======================= #

def test_predict_single_with_prior_with_return_scores():
    """Predict a single observation sequence with a prior and returned scores"""
    prediction = hmm_clf.predict(x, prior=True, return_scores=True)
    assert isinstance(prediction, tuple)
    assert isinstance(prediction[0], str)
    assert isinstance(prediction[1], list)
    assert isinstance(prediction[1][0], tuple)
    assert isinstance(prediction[1][0][0], str)
    assert isinstance(prediction[1][0][1], float)
    assert len(prediction[1]) == 5

def test_predict_single_with_prior_no_return_scores():
    """Predict a single observation sequence with a prior and no returned scores"""
    prediction = hmm_clf.predict(x, prior=True, return_scores=False)
    assert isinstance(prediction, str)

def test_predict_single_no_prior_with_return_scores():
    """Predict a single observation sequence with no prior and returned scores"""
    prediction = hmm_clf.predict(x, prior=False, return_scores=True)
    assert isinstance(prediction, tuple)
    assert isinstance(prediction[0], str)
    assert isinstance(prediction[1], list)
    assert isinstance(prediction[1][0], tuple)
    assert isinstance(prediction[1][0][0], str)
    assert isinstance(prediction[1][0][1], float)
    assert len(prediction[1]) == 5

def test_predict_single_no_prior_no_return_scores():
    """Predict a single observation sequence with no prior and no returned scores"""
    prediction = hmm_clf.predict(x, prior=False, return_scores=False)
    assert isinstance(prediction, str)

def test_predict_multiple_with_prior_with_return_scores():
    """Predict multiple observation sequences with a prior and returned scores"""
    predictions = hmm_clf.predict(X, prior=True, return_scores=True)
    assert isinstance(predictions, list)
    assert len(predictions) == 3
    # First prediction
    assert isinstance(predictions[0], tuple)
    assert isinstance(predictions[0][0], str)
    assert isinstance(predictions[0][1], list)
    assert len(predictions[0][1]) == 5
    # First score
    assert isinstance(predictions[0][1][0], tuple)
    assert isinstance(predictions[0][1][0][0], str)
    assert isinstance(predictions[0][1][0][1], float)

def test_predict_multiple_with_prior_no_return_scores():
    """Predict multiple observation sequences with a prior and no returned scores"""
    predictions = hmm_clf.predict(X, prior=True, return_scores=False)
    assert isinstance(predictions, list)
    assert len(predictions) == 3

def test_predict_multiple_no_prior_with_return_scores():
    """Predict multiple observation sequences with no prior and returned scores"""
    predictions = hmm_clf.predict(X, prior=False, return_scores=True)
    assert isinstance(predictions, list)
    assert len(predictions) == 3
    # First prediction
    assert isinstance(predictions[0], tuple)
    assert isinstance(predictions[0][0], str)
    assert isinstance(predictions[0][1], list)
    assert len(predictions[0][1]) == 5
    # First score
    assert isinstance(predictions[0][1][0], tuple)
    assert isinstance(predictions[0][1][0][0], str)
    assert isinstance(predictions[0][1][0][1], float)

def test_predict_single_no_prior_no_return_scores():
    """Predict multiple observation sequences with no prior and no returned scores"""
    predictions = hmm_clf.predict(X, prior=False, return_scores=False)
    assert isinstance(predictions, list)
    assert len(predictions) == 3

# ======================== #
# HMMClassifier.evaluate() #
# ======================== #

def test_evaluate_with_prior_with_labels():
    """Evaluate with a prior and confusion matrix labels"""
    acc, cm = hmm_clf.evaluate(X, Y, prior=True, labels=['c0', 'c1', 'c2', 'c3', 'c4'])
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)
    assert cm.shape == (5, 5)

def test_evaluate_with_prior_no_labels():
    """Evaluate with a prior and no confusion matrix labels"""
    acc, cm = hmm_clf.evaluate(X, Y, prior=True, labels=None)
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)

def test_evaluate_no_prior_with_labels():
    """Evaluate with no prior and confusion matrix labels"""
    acc, cm = hmm_clf.evaluate(X, Y, prior=False, labels=['c0', 'c1', 'c2', 'c3', 'c4'])
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)
    assert cm.shape == (5, 5)

def test_evaluate_no_prior_no_labels():
    """Evaluate with no prior and no confusion matrix labels"""
    acc, cm = hmm_clf.evaluate(X, Y, prior=False, labels=None)
    assert isinstance(acc, float)
    assert isinstance(cm, np.ndarray)


import numpy as np
from .methods import (
    _trim_zeros, _center, _standardize, _downsample, _fft, _filtrate
)
from ..internals import _Validator

class Preprocess:
    """Efficiently applies multiple preprocessing transformations to the provided input observation sequence(s)."""

    def __init__(self):
        self._transforms = []
        self._val = _Validator()

    def trim_zeros(self):
        """Trim zero-observations from the input observation sequence(s)."""
        self._transforms.append((_trim_zeros, {}))

    def center(self):
        """Centers an observation sequence (or multiple sequences) by centering observations around the mean."""
        self._transforms.append((_center, {}))

    def standardize(self):
        """Standardizes an observation sequence (or multiple sequences) by transforming observations
        so that they have zero mean and unit variance."""
        self._transforms.append((_standardize, {}))

    def downsample(self, n, method='decimate'):
        """Downsamples an observation sequence (or multiple sequences) by either:

        - Decimating the next :math:`n-1` observations
        - Averaging the current observation with the next :math:`n-1` observations

        Parameters
        ----------
        n: int
            Downsample factor.

        method: {'decimate', 'average'}
            The downsampling method.
        """
        self._val.restricted_integer(n, lambda x: x > 1, desc='downsample factor', expected='greater than one')
        self._val.one_of(method, ['decimate', 'average'], desc='downsampling method')
        self._transforms.append((_downsample, {'n': n, 'method': method}))

    def fft(self):
        """Applies a Discrete Fourier Transform to the input observation sequence(s)."""
        self._transforms.append((_fft, {}))

    def filtrate(self, n, method='median'):
        """Applies a median or mean filter to the input observation sequence(s).

        Parameters
        ----------
        n: int
            Window size.

        method: {'median', 'mean'}
            The filtering method.
        """
        self._val.restricted_integer(n, lambda x: x > 1, desc='window size', expected='greater than one')
        self._val.one_of(method, ['median', 'mean'], desc='filtering method')
        self._transforms.append((_filtrate, {'n': n, 'method': method}))

    def transform(self, X):
        """Applies the preprocessing transformations to the provided input observation sequence(s).

        Parameters
        ----------
        X: numpy.ndarray or List[numpy.ndarray]
            An individual observation sequence or a list of multiple observation sequences.

        Returns
        -------
        transformed: numpy.ndarray or List[numpy.ndarray]
            The input observation sequence(s) with preprocessing transformations applied in order.
        """
        self._val.observation_sequences(X, allow_single=True)

        X_transform = X
        for transform, kwargs in self._transforms:
            if transform == _downsample:
                if isinstance(X_transform, np.ndarray):
                    self._val.restricted_integer(kwargs['n'], lambda x: x <= len(X_transform),
                        desc='downsample factor', expected='no greater than the number of frames')
                else:
                    self._val.restricted_integer(kwargs['n'], lambda x: x <= min(len(x) for x in X_transform),
                        desc='downsample factor', expected='no greater than the number of frames in the shortest sequence')
            elif transform == _filtrate:
                if isinstance(X, np.ndarray):
                    self._val.restricted_integer(kwargs['n'], lambda x: x <= len(X),
                        desc='window size', expected='no greater than the number of frames')
                else:
                    self._val.restricted_integer(kwargs['n'], lambda x: x <= min(len(x) for x in X),
                        desc='window size', expected='no greater than the number of frames in the shortest sequence')
            X_transform = transform(X_transform, **kwargs)
        return X_transform

    def summary(self):
        """Displays an ordered summary of the preprocessing transformations."""
        if len(self._transforms) == 0:
            raise RuntimeError('At least one preprocessing transformation is required')

        steps = []

        for i, (transform, kwargs) in enumerate(self._transforms):
            idx = i + 1
            if transform == _center:
                steps.append(('{}. Centering'.format(idx), None))
            elif transform == _standardize:
                steps.append(('{}. Standardization'.format(idx), None))
            elif transform == _downsample:
                header = 'Decimation' if kwargs['method'] == 'decimate' else 'Averaging'
                steps.append((
                    '{}. Downsampling:'.format(idx),
                    '   {} with downsample factor (n={})'.format(header, kwargs['n'])
                ))
            elif transform == _fft:
                steps.append(('{}. Discrete Fourier Transform'.format(idx), None))
            elif transform == _filtrate:
                steps.append((
                    '{}. Filtering:'.format(idx),
                    '   {} filter with window size (n={})'.format(kwargs['method'].capitalize(), kwargs['n'])
                ))
            elif transform == _trim_zeros:
                steps.append(('{}. Zero-trimming'.format(idx), None))

        title = 'Preprocessing summary:'
        length = max(max(len(h), 0 if b is None else len(b)) for h, b in steps)
        length = len(title) if length < len(title) else length

        print(title.center(length, ' '))
        print('=' * length)
        for i, (head, body) in enumerate(steps):
            print(head)
            if body is not None:
                print(body)
            if i != len(steps) - 1:
                print('-' * length)
        print('=' * length)


import pytest
import numpy as np
from sequentia.preprocessing import (
    Preprocess,
    trim_zeros, downsample, center, standardize, fft, filtrate,
    _trim_zeros, _downsample, _center, _standardize, _fft, _filtrate
)
from ...support import assert_equal, assert_all_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

# Sample data
X = rng.random((7, 2))
Xs = [i * rng.random((3 * i, 2)) for i in range(1, 4)]

# Zero-trimming preprocessor
trim = Preprocess()
trim.trim_zeros()

# Centering preprocessor
cent = Preprocess()
cent.center()

# Standardizing preprocessor
standard = Preprocess()
standard.standardize()

# Discrete Fourier Transform preprocessor
fourier = Preprocess()
fourier.fft()

# Downsampling preprocessor
down = Preprocess()
down_kwargs = {'n': 3, 'method': 'decimate'}
down.downsample(**down_kwargs)

# Filtering preprocessor
filt = Preprocess()
filt_kwargs = {'n': 3, 'method': 'median'}
filt.filtrate(**filt_kwargs)

# Combined preprocessor
combined = Preprocess()
combined.trim_zeros()
combined.center()
combined.standardize()
combined.filtrate(**filt_kwargs)
combined.downsample(**down_kwargs)
combined.fft()

# ======================= #
# Preprocess.trim_zeros() #
# ======================= #

def test_trim_zeros_adds_transform():
    """Applying a single zero-trimming transformation"""
    assert len(trim._transforms) == 1
    assert trim._transforms[0] == (_trim_zeros, {})

def test_trim_zeros_single():
    """Applying zero-trimming to a single observation sequence"""
    assert_equal(trim.transform(X), trim_zeros(X))

def test_trim_zeros_multiple():
    """Applying zero-trimming to multiple observation sequences"""
    assert_all_equal(trim.transform(Xs), trim_zeros(Xs))

def test_trim_zeros_summary(capsys):
    """Summary of a zero-trimming transformation"""
    trim.summary()
    assert capsys.readouterr().out == (
        'Preprocessing summary:\n'
        '======================\n'
        '1. Zero-trimming\n'
        '======================\n'
    )

# =================== #
# Preprocess.center() #
# =================== #

def test_center_adds_transform():
    """Applying a single centering transformation"""
    assert len(cent._transforms) == 1
    assert cent._transforms[0] == (_center, {})

def test_center_single():
    """Applying centering to a single observation sequence"""
    assert_equal(cent.transform(X), center(X))

def test_center_multiple():
    """Applying centering to multiple observation sequences"""
    assert_all_equal(cent.transform(Xs), center(Xs))

def test_center_summary(capsys):
    """Summary of a centering transformation"""
    cent.summary()
    assert capsys.readouterr().out == (
        'Preprocessing summary:\n'
        '======================\n'
        '1. Centering\n'
        '======================\n'
    )

# ======================== #
# Preprocess.standardize() #
# ======================== #

def test_standardize_adds_transform():
    """Applying a single standardizing transformation"""
    assert len(standard._transforms) == 1
    assert standard._transforms[0] == (_standardize, {})

def test_standardize_single():
    """Applying standardization to a single observation sequence"""
    assert_equal(standard.transform(X), standardize(X))

def test_standardize_multiple():
    """Applying standardization to multiple observation sequences"""
    assert_all_equal(standard.transform(Xs), standardize(Xs))

def test_standardize_summary(capsys):
    """Summary of a standardizing transformation"""
    standard.summary()
    assert capsys.readouterr().out == (
        'Preprocessing summary:\n'
        '======================\n'
        '1. Standardization\n'
        '======================\n'
    )

# ================ #
# Preprocess.fft() #
# ================ #

def test_fft_adds_transform():
    """Applying a single discrete fourier transformation"""
    assert len(fourier._transforms) == 1
    assert fourier._transforms[0] == (_fft, {})

def test_fft_single():
    """Applying discrete fourier transformation to a single observation sequence"""
    assert_equal(fourier.transform(X), fft(X))

def test_fft_multiple():
    """Applying discrete fourier transformation to multiple observation sequences"""
    assert_all_equal(fourier.transform(Xs), fft(Xs))

def test_fft_summary(capsys):
    """Summary of a discrete fourier transformation"""
    fourier.summary()
    assert capsys.readouterr().out == (
        '    Preprocessing summary:   \n'
        '=============================\n'
        '1. Discrete Fourier Transform\n'
        '=============================\n'
    )

# ======================= #
# Preprocess.downsample() #
# ======================= #

def test_downsample_adds_transform():
    """Applying a single downsampling transformation"""
    assert len(down._transforms) == 1
    assert down._transforms[0] == (_downsample, down_kwargs)

def test_downsample_single():
    """Applying downsampling to a single observation sequence"""
    assert_equal(down.transform(X), downsample(X, **down_kwargs))

def test_downsample_multiple():
    """Applying downsampling to multiple observation sequences"""
    assert_all_equal(down.transform(Xs), downsample(Xs, **down_kwargs))

def test_downsample_summary(capsys):
    """Summary of a downsampling transformation"""
    down.summary()
    assert capsys.readouterr().out == (
        '          Preprocessing summary:          \n'
        '==========================================\n'
        '1. Downsampling:\n'
        '   Decimation with downsample factor (n=3)\n'
        '==========================================\n'
    )

# ===================== #
# Preprocess.filtrate() #
# ===================== #

def test_filtrate_adds_transform():
    """Applying a single filtering transformation"""
    assert len(filt._transforms) == 1
    assert filt._transforms[0] == (_filtrate, filt_kwargs)

def test_filtrate_single():
    """Applying filtering to a single observation sequence"""
    assert_equal(filt.transform(X), filtrate(X, **filt_kwargs))

def test_filtrate_multiple():
    """Applying filtering to multiple observation sequences"""
    assert_all_equal(filt.transform(Xs), filtrate(Xs, **filt_kwargs))

def test_filtrate_summary(capsys):
    """Summary of a filtering transformation"""
    filt.summary()
    assert capsys.readouterr().out == (
        '         Preprocessing summary:        \n'
        '=======================================\n'
        '1. Filtering:\n'
        '   Median filter with window size (n=3)\n'
        '=======================================\n'
    )

# ======================== #
# Combined transformations #
# ======================== #

def test_combined_adds_transforms():
    """Applying multiple filtering transformations"""
    assert len(combined._transforms) == 6
    assert combined._transforms == [
        (_trim_zeros, {}),
        (_center, {}),
        (_standardize, {}),
        (_filtrate, filt_kwargs),
        (_downsample, down_kwargs),
        (_fft, {})
    ]

def test_combined_single():
    """Applying combined transformations to a single observation sequence"""
    X_pre = X
    X_pre = trim_zeros(X_pre)
    X_pre = center(X_pre)
    X_pre = standardize(X_pre)
    X_pre = filtrate(X_pre, **filt_kwargs)
    X_pre = downsample(X_pre, **down_kwargs)
    X_pre = fft(X_pre)
    assert_equal(combined.transform(X), X_pre)

def test_combined_multiple():
    """Applying combined transformations to multiple observation sequences"""
    Xs_pre = Xs
    Xs_pre = trim_zeros(Xs_pre)
    Xs_pre = center(Xs_pre)
    Xs_pre = standardize(Xs_pre)
    Xs_pre = filtrate(Xs_pre, **filt_kwargs)
    Xs_pre = downsample(Xs_pre, **down_kwargs)
    Xs_pre = fft(Xs_pre)
    assert_all_equal(combined.transform(Xs), Xs_pre)

def test_combined_summary(capsys):
    """Summary with combined transformations applied"""
    combined.summary()
    assert capsys.readouterr().out == (
        '          Preprocessing summary:          \n'
        '==========================================\n'
        '1. Zero-trimming\n'
        '------------------------------------------\n'
        '2. Centering\n'
        '------------------------------------------\n'
        '3. Standardization\n'
        '------------------------------------------\n'
        '4. Filtering:\n'
        '   Median filter with window size (n=3)\n'
        '------------------------------------------\n'
        '5. Downsampling:\n'
        '   Decimation with downsample factor (n=3)\n'
        '------------------------------------------\n'
        '6. Discrete Fourier Transform\n'
        '==========================================\n'
    )

# ==================== #
# Preprocess.summary() #
# ==================== #

def test_empty_summary():
    """Summary without any transformations applied"""
    with pytest.raises(RuntimeError) as e:
        Preprocess().summary()
    assert str(e.value) == 'At least one preprocessing transformation is required'


import numpy as np
from .hmm import HMM
from sklearn.metrics import confusion_matrix
from ...internals import _Validator

class HMMClassifier:
    """An ensemble classifier that combines individual :class:`~HMM` objects, which model isolated sequences from different classes."""

    def __init__(self):
        self._val = _Validator()

    def fit(self, models):
        """Fits the ensemble classifier with a collection of :class:`~HMM` objects.

        Parameters
        ----------
        models: List[HMM] or Dict[Any, HMM]
            A collection of :class:`~HMM` objects to use for classification.
        """
        if isinstance(models, list):
            if not all(isinstance(model, HMM) for model in models):
                raise TypeError('Expected all models to be HMM objects')
        elif isinstance(models, dict):
            values = list(models.values())
            if not all(isinstance(model, HMM) for model in values):
                raise TypeError('Expected all models to be HMM objects')
            models = values
        else:
            raise TypeError('Expected models to be a list or dict of HMM objects')

        if len(models) > 0:
            self._models = models
        else:
            raise RuntimeError('Must fit the classifier with at least one HMM')

    def predict(self, X, prior=True, return_scores=False):
        """Predicts the label for an observation sequence (or multiple sequences) according to maximum likelihood or posterior scores.

        Parameters
        ----------
        X: numpy.ndarray or List[numpy.ndarray]
            An individual observation sequence or a list of multiple observation sequences.

        prior: bool
            Whether to calculate a prior for each model and perform MAP estimation by scoring with
            the joint probability (or un-normalized posterior) :math:`\mathbb{P}(O, \lambda_c)=\mathbb{P}(O|\lambda_c)\mathbb{P}(\lambda_c)`.

            If this parameter is set to false, then the negative log likelihoods
            :math:`\mathbb{P}(O|\lambda_c)` generated from the models' :func:`~HMM.forward` function are used.

        return_scores: bool
            Whether to return the scores of each model on the observation sequence(s).

        Returns
        -------
        prediction(s): str or List[str]
            The predicted label(s) for the observation sequence(s).

            If ``return_scores`` is true, then for each observation sequence, a tuple `(label, scores)` is returned for each label,
            consisting of the `scores` of each HMM and the `label` of the HMM with the best score.
        """
        self._val.boolean(prior, desc='prior')
        self._val.boolean(return_scores, desc='return_scores')
        self._val.observation_sequences(X, allow_single=True)

        try:
            self._models
        except AttributeError as e:
            raise AttributeError('The classifier needs to be fitted before predictions are made') from e

        total_seqs = sum(model.n_seqs for model in self._models)

        if isinstance(X, np.ndarray): # Single observation sequence
            scores = [(model.label, model.forward(X) - np.log(model.n_seqs / total_seqs) * prior) for model in self._models]
            best = min(scores, key=lambda x: x[1])
            return (best[0], scores) if return_scores else best[0]
        else: # Multiple observation sequences
            predictions = []
            for x in X:
                scores = [(model.label, model.forward(x) - np.log(model.n_seqs / total_seqs) * prior) for model in self._models]
                best = min(scores, key=lambda x: x[1])
                predictions.append((best[0], scores) if return_scores else best[0])
            return predictions

    def evaluate(self, X, y, prior=True, labels=None):
        """Evaluates the performance of the classifier on a batch of observation sequences and their labels.

        Parameters
        ----------
        X: List[numpy.ndarray]
            A list of multiple observation sequences.

        y: List[str]
            A list of labels for the observation sequences.

        prior: bool
            Whether to calculate a prior for each model and perform MAP estimation by scoring with
            the joint probability (or un-normalized posterior) :math:`\mathbb{P}(O, \lambda_c)=\mathbb{P}(O|\lambda_c)\mathbb{P}(\lambda_c)`.

            If this parameter is set to false, then the negative log likelihoods
            :math:`\mathbb{P}(O|\lambda_c)` generated from the models' :func:`~HMM.forward` function are used.

        labels: List[str]
            A list of labels for ordering the axes of the confusion matrix.

        Returns
        -------
        accuracy: float
            The categorical accuracy of the classifier on the observation sequences.

        confusion: numpy.ndarray
            The confusion matrix representing the discrepancy between predicted and actual labels.
        """
        self._val.observation_sequences_and_labels(X, y)
        self._val.boolean(prior, desc='prior')

        if labels is not None:
            self._val.list_of_strings(labels, desc='confusion matrix labels')

        # Classify each observation sequence and calculate confusion matrix
        predictions = self.predict(X, prior, return_scores=False)
        cm = confusion_matrix(y, predictions, labels=labels)

        return np.sum(np.diag(cm)) / np.sum(cm), cm


import numpy as np

class _Validator:
    """Performs internal validations on various input types."""

    def observation_sequences(self, X, allow_single=False):
        """Validates observation sequence(s).

        Parameters
        ----------
        X: numpy.ndarray or List[numpy.ndarray]
            An individual observation sequence or a list of multiple observation sequences.

        allow_single: bool
            Whether to allow an individual observation sequence.

        Returns
        -------
        X: numpy.ndarray or List[numpy.ndarray]
            The original input observation sequence(s) if valid.
        """
        if isinstance(X, (list, np.ndarray) if allow_single else list):
            if isinstance(X, list):
                if not all(isinstance(sequence, np.ndarray) for sequence in X):
                    raise TypeError('Each observation sequence must be a numpy.ndarray')
                if not all(sequence.ndim == 2 for sequence in X):
                    raise ValueError('Each observation sequence must be two-dimensional')
                if not all(sequence.shape[1] == X[0].shape[1] for sequence in X):
                    raise ValueError('Each observation sequence must have the same dimensionality')
            elif isinstance(X, np.ndarray):
                if not X.ndim == 2:
                    raise ValueError('Observation sequence must be two-dimensional')
        else:
            if allow_single:
                raise TypeError('Expected an individual observation sequence or a list of multiple observation sequences, each of type numpy.ndarray')
            else:
                raise TypeError('Expected a list of observation sequences, each of type numpy.ndarray')
        return X

    def observation_sequences_and_labels(self, X, y):
        """Validates observation sequences and corresponding labels.

        Parameters
        ----------
        X: List[numpy.ndarray]
            A list of multiple observation sequences.

        y: List[str]
            A list of labels for the observation sequences.

        Returns
        -------
        X: List[numpy.ndarray]
            The original input observation sequences if valid.

        y: List[str]
            The original input labels if valid.
        """
        self.observation_sequences(X, allow_single=False)
        self.list_of_strings(y, desc='labels')
        if not len(X) == len(y):
            raise ValueError('Expected the same number of observation sequences and labels')
        return X, y

    def integer(self, item, desc):
        """Validates an integer.

        Parameters
        ----------
        item: int
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: int
            The original input item if valid.
        """
        if not isinstance(item, int):
            raise TypeError("Expected {} to be an integer".format(desc))
        return item

    def string(self, item, desc):
        """Validates a string.

        Parameters
        ----------
        item: str
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: str
            The original input item if valid.
        """
        if not isinstance(item, str):
            raise TypeError("Expected {} to be a string".format(desc))
        return item

    def boolean(self, item, desc):
        """Validates a boolean.

        Parameters
        ----------
        item: bool
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: bool
            The original input item if valid.
        """
        if not isinstance(item, bool):
            raise TypeError("Expected {} to be a boolean".format(desc))
        return item

    def one_of(self, item, items, desc):
        """Validates that an item is one of some permitted values.

        Parameters
        ----------
        item: Any
            The item to validate.

        items: List[Any]
            The list of permitted values to check against.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: Any
            The original input item if valid.
        """
        if not item in items:
            raise ValueError('Expected {} to be one of {}'.format(desc, items))
        return item

    def restricted_integer(self, item, condition, desc, expected):
        """Validates an integer and checks that it satisfies some condition.

        Parameters
        ----------
        item: int
            The item to validate.

        condition: lambda
            A condition to check the item against.

        desc: str
            A description of the item being validated.

        expected: str
            A description of the condition, or expected value.

        Returns
        -------
        item: int
            The original input item if valid.
        """
        if isinstance(item, int):
            if not condition(item):
                raise ValueError('Expected {} to be {}'.format(desc, expected))
        else:
            raise TypeError("Expected {} to be an integer".format(desc))
        return item

    def restricted_float(self, item, condition, desc, expected):
        """Validates a float and checks that it satisfies some condition.

        Parameters
        ----------
        item: float
            The item to validate.

        condition: lambda
            A condition to check the item against.

        desc: str
            A description of the item being validated.

        expected: str
            A description of the condition, or expected value.

        Returns
        -------
        item: float
            The original input item if valid.
        """
        if isinstance(item, float):
            if not condition(item):
                raise ValueError('Expected {} to be {}'.format(desc, expected))
        else:
            raise TypeError("Expected {} to be a float".format(desc))
        return item

    def list_of_strings(self, items, desc):
        """Validates a list and checks that it consists entirely of strings.

        Parameters
        ----------
        items: List[str]
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        items: List[str]
            The original input items if valid.
        """
        if isinstance(items, list):
            if not all(isinstance(item, str) for item in items):
                raise ValueError('Expected all {} to be strings'.format(desc))
        else:
            raise TypeError('Expected {} to be a list of strings'.format(desc))
        return items

    def random_state(self, state):
        """Validates a random state object or seed.

        Parameters
        ----------
        state: None, int, numpy.random.RandomState
            A random state object or seed.

        Returns
        -------
        state: numpy.random.RandomState
            A random state object.
        """
        if state is None:
            return np.random.RandomState(seed=0)
        elif isinstance(state, int):
            return np.random.RandomState(seed=state)
        elif isinstance(state, np.random.RandomState):
            return state
        else:
            raise TypeError('Expected random state to be of type: None, int, or numpy.random.RandomState')


import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal

def assert_equal(a, b):
    assert_allclose(a, b, rtol=1e-5)

def assert_not_equal(a, b):
    assert not np.allclose(a, b, rtol=1e-5)

def assert_all_equal(A, B):
    for a, b in zip(A, B):
        assert_equal(a, b)

def assert_distribution(a):
    if a.ndim == 1:
        assert_almost_equal(a.sum(), 1., decimal=5)
    elif a.ndim == 2:
        assert_almost_equal(a.sum(axis=1), np.ones(len(a)))